@@ -691,7 +691,8 @@ QVariant KDescendantsProxyModel::headerData(int section, Qt::Orientation orienta
         return QVariant();
     }
 
-    return QAbstractProxyModel::headerData(section, orientation, role);
+    // Here is safe to do sourceModel()->headerData, as in this proxy we neither filter out nor reorder columns
+    return sourceModel()->headerData(section, orientation, role);
 }
 
 Qt::ItemFlags KDescendantsProxyModel::flags(const QModelIndex &index) const
@@ -780,7 +781,9 @@ void KDescendantsProxyModelPrivate::sourceRowsInserted(const QModelIndex &parent
 
     if (rowCount == difference) {
         const QModelIndex index = q->mapFromSource(parent);
-        Q_EMIT q->dataChanged(index, index, {KDescendantsProxyModel::ExpandableRole, KDescendantsProxyModel::ExpandedRole, KDescendantsProxyModel::HasSiblingsRole});
+        if (parent.isValid()) {
+            Q_EMIT q->dataChanged(index, index, {KDescendantsProxyModel::ExpandableRole, KDescendantsProxyModel::ExpandedRole, KDescendantsProxyModel::HasSiblingsRole});
+        }
         // @p parent was not a parent before.
         m_pendingParents.append(parent);
         scheduleProcessPendingParents();
@@ -880,12 +883,15 @@ void KDescendantsProxyModelPrivate::sourceRowsInserted(const QModelIndex &parent
 
     q->endInsertRows();
     scheduleProcessPendingParents();
-    const QModelIndex index = q->mapFromSource(parent);
-    Q_EMIT q->dataChanged(index, index, {KDescendantsProxyModel::ExpandableRole, KDescendantsProxyModel::ExpandedRole, KDescendantsProxyModel::HasSiblingsRole});
+    if (parent.isValid()) {
+        const QModelIndex index = q->mapFromSource(parent);
+        Q_EMIT q->dataChanged(index, index, {KDescendantsProxyModel::ExpandableRole, KDescendantsProxyModel::ExpandedRole, KDescendantsProxyModel::HasSiblingsRole});
+    }
 
     if (start > 0) {
         notifyhasSiblings(q->sourceModel()->index(start - 1, 0, parent));
     }
+
 }
 
 void KDescendantsProxyModelPrivate::sourceRowsAboutToBeRemoved(const QModelIndex &parent, int start, int end)
@@ -938,8 +944,10 @@ void KDescendantsProxyModelPrivate::sourceRowsRemoved(const QModelIndex &parent,
     Q_UNUSED(end)
 
     if (!q->isSourceIndexExpanded(parent) || !q->isSourceIndexVisible(parent)) {
-        const QModelIndex index = q->mapFromSource(parent);
-        Q_EMIT q->dataChanged(index, index, {KDescendantsProxyModel::ExpandableRole, KDescendantsProxyModel::ExpandedRole});
+        if (parent.isValid()) {
+            const QModelIndex index = q->mapFromSource(parent);
+            Q_EMIT q->dataChanged(index, index, {KDescendantsProxyModel::ExpandableRole, KDescendantsProxyModel::ExpandedRole});
+        }
         return;
     }
 
@@ -971,8 +979,10 @@ void KDescendantsProxyModelPrivate::sourceRowsRemoved(const QModelIndex &parent,
 
     if (rowCount != start || rowCount == 0) {
         q->endRemoveRows();
-        const QModelIndex index = q->mapFromSource(parent);
-        Q_EMIT q->dataChanged(index, index, {KDescendantsProxyModel::ExpandableRole, KDescendantsProxyModel::ExpandedRole});
+        if (parent.isValid()) {
+            const QModelIndex index = q->mapFromSource(parent);
+            Q_EMIT q->dataChanged(index, index, {KDescendantsProxyModel::ExpandableRole, KDescendantsProxyModel::ExpandedRole});
+        }
         if (start > 0) {
             notifyhasSiblings(q->sourceModel()->index(start - 1, 0, parent));
         }
@@ -1093,15 +1103,20 @@ void KDescendantsProxyModelPrivate::sourceRowsRemoved(const QModelIndex &parent,
     m_mapping.insert(newEnd, proxyRow + newEnd.row());
     q->endRemoveRows();
 
-    const QModelIndex index = q->mapFromSource(parent);
-    Q_EMIT q->dataChanged(index, index, {KDescendantsProxyModel::ExpandableRole});
+    if (parent.isValid()) {
+        const QModelIndex oindex = q->mapFromSource(parent);
+        QVector<int> rolesChanged({KDescendantsProxyModel::ExpandableRole});
 
-    if (!q->sourceModel()->hasChildren(parent)) {
-        Q_EMIT q->dataChanged(index, index, {KDescendantsProxyModel::ExpandedRole});
-    } else if (q->sourceModel()->rowCount(parent) <= start) {
-        const QModelIndex index = q->mapFromSource(q->sourceModel()->index(q->sourceModel()->rowCount(parent) - 1, 0, parent));
-        Q_EMIT q->dataChanged(index, index, {KDescendantsProxyModel::ExpandedRole});
+        if (!q->sourceModel()->hasChildren(parent)) {
+            rolesChanged << KDescendantsProxyModel::ExpandedRole;
+        } else if (q->sourceModel()->rowCount(parent) <= start) {
+            const QModelIndex index = q->mapFromSource(q->sourceModel()->index(q->sourceModel()->rowCount(parent) - 1, 0, parent));
+            Q_EMIT q->dataChanged(index, index, {KDescendantsProxyModel::ExpandedRole});
+        }
+
+        Q_EMIT q->dataChanged(oindex, oindex, rolesChanged);
     }
+
     if (start > 0) {
         notifyhasSiblings(q->sourceModel()->index(start - 1, 0, parent));
     }
@@ -1112,6 +1127,21 @@ void KDescendantsProxyModelPrivate::sourceRowsAboutToBeMoved(const QModelIndex &
     Q_Q(KDescendantsProxyModel);
 
     Q_UNUSED(destStart)
+
+    if (q->isSourceIndexExpanded(srcParent) && q->isSourceIndexVisible(srcParent) &&
+        (!q->isSourceIndexExpanded(destParent) || !q->isSourceIndexVisible(destParent))) {
+        const QModelIndex proxySrcParent = q->mapFromSource(srcParent);
+        const int proxyParentRow = proxySrcParent.isValid() ? proxySrcParent.row() : 0;
+        q->beginRemoveRows(QModelIndex(), proxyParentRow + srcStart, proxyParentRow + srcEnd);
+
+    } else if ((!q->isSourceIndexExpanded(srcParent) || !q->isSourceIndexVisible(srcParent)) &&
+        q->isSourceIndexExpanded(destParent) && q->isSourceIndexVisible(destParent)) {
+        const QModelIndex proxyDestParent = q->mapFromSource(srcParent);
+        const int proxyParentRow = proxyDestParent.isValid() ? proxyDestParent.row() : 0;
+
+        q->beginInsertRows(QModelIndex(), proxyParentRow + destStart, proxyParentRow + destStart + (srcEnd - srcStart));
+    }
+
     sourceLayoutAboutToBeChanged();
 }
 
@@ -1124,6 +1154,14 @@ void KDescendantsProxyModelPrivate::sourceRowsMoved(const QModelIndex &srcParent
     Q_UNUSED(srcEnd)
     Q_UNUSED(destParent)
     Q_UNUSED(destStart)
+
+    if (q->isSourceIndexExpanded(srcParent) && q->isSourceIndexVisible(srcParent) &&
+        (!q->isSourceIndexExpanded(destParent) || !q->isSourceIndexVisible(destParent))) {
+        q->endRemoveRows();
+    } else if (!q->isSourceIndexExpanded(srcParent) && q->isSourceIndexExpanded(destParent)) {
+        q->endInsertRows();
+    }
+
     sourceLayoutChanged();
 
     const QModelIndex index1 = q->mapFromSource(srcParent);
